"""
This module inserts the magic key and allows to turn it on or off.

Examples:
    >>> import magickey
    >>> from .examples.person import Person

    >>> merlin = Person("Myrddin Wyllt", 42, "Caledonia") 


    >>> # Using the default iPython matching engine
    >>> magickey.turn_on(merlin, magic_type = None)
    >>> @`merlin.name()`
    Myrddin Wyllt

    >>> @merlin What is your first name?
    Invalid .. .  # TODO add actual error

    >>>
    >>> # Using the default Whoosh sting matching engine
    >>> magickey.turn_on(merlin, magic_type = False)
    >>> @merlin What is your first name?
    Myrddin Wyllt

    >>>
    >>> # Using the intellegence engine
    >>> magickey.turn_on(merlin, magic_type = True)
    
    >>> @`merlin.name()`
    Myrddin Wyllt

    >>> @merlin Please, can you remind me, what is your first name?  It's M... ?
    It's Merlin. People also call me Merlinus and Myrddin. Please, feel free to call me Merlin. It is easier this way in English.

    >>> magickey.turn_off(merlin)      # By default, this backs up the episode and consolidates the memory
"""

def turn_on(object, magic_type=None):
    """
    Activates the connector between the python interpreter and intellegence engine. Uses no magic by default.
    :param magic_type: Can be None, False or True. Specifies the type of magic to use in string matching.
    """

    match magic_type:
        case None:              # exact sting matching
            pass

        case False:              # whoosh engine sting matching
            pass

        case None:              # intellegence engine sting matching
            pass


import magickey.magics

# Load the extension when this module is loaded.
magics.load_ipython_extension(get_ipython())





# This code can be put in any Python module, it does not require IPython
# itself to be running already.  It only creates the magics subclass but
# doesn't instantiate it yet.
from __future__ import print_function

import IPython.core.magic
from IPython.display import display, Javascript, Markdown, Code
from IPython.core.magic import (Magics, magics_class, line_magic,
                                cell_magic, line_cell_magic)
import ast      # AST is also magic, right?
from parsimonious.nodes import NodeVisitor      # And so are PEGs!
from parsimonious.grammar import Grammar

import openai, os, getpass, boto3
from notebook.utils import to_api_path

# load prompt.txt from module resources with importlib
import importlib.resources as pkg_resources
prompt_txt = pkg_resources.read_text(__package__, 'prompt.txt')

print("prompt:", prompt_txt)



@magics_class
class NoMagic(Magics):
    """" This implementation uses exact string matching"""
    pass


@magics_class
class FalseMagic(Magics):
    """" This implementation uses whoosh search backend"""
    pass


@magics_class
class TrueMagic(Magics):
    """" This implementation uses AI backend"""

    def __init__(self, shell, **kwargs):
        super().__init__(shell, **kwargs)   
        
        # Initialize inference engine
        openai.api_key = open("/etc/.openai.merlin.key", 'rt').read().strip()
        # os.getenv("OPENAI_API_KEY")    

     

    @line_magic
    def asterisk(self, line):
        "User prompt"

        name, input = line[5:-4].split(':%* ', maxsplit = 1)
        name = name[0].upper() + name[1:]
        prompt = prompt_txt + '\n\n\n' + name + ': ' + input + 'Arthur: '

        
        print(prompt)


        response = openai.Completion.create(
            model="text-davinci-003",
            prompt=prompt,
            temperature=0.6,
            max_tokens=100,
        )

        # Note, it is fine that the response could contain prediction of responces for other parts of the system.
        # It doesn't mean that these predictions will be used, the prediction then can be compared with the
        # actual response and the AI can be notified and, if beneficial, finetuned, to improve its predictions!

        # As human we are similar in that and we predict the next word in the sentence. When the word is not what we
        # expect, we are surprised. We can use this to our advantage. We can use the surprise to improve the AI.
        
        # When the actual response is different from the predicted one, we'll tag it with #surprise.

        arthur = response.choices[0].text
        # add_response_cell(arthur)

        #print("Prompting. Full access to the main IPython object:", self.shell)
        #print("Variables in the user namespace:", list(self.shell.user_ns.keys()))
        #display(Markdown(text))

        lines = arthur_to_python(arthur)
        add_code_cell("\n".join(lines))

        add_prompt_cell()
        return lines


        
        # make output markdown
        # https://stackoverflow.com/questions/47818822/can-i-define-a-custom-cell-magic-in-ipython



    @cell_magic
    def thread(self, line, cell):
        "This allows to create a temporary thread"
        # http://ipython.org/ipython-doc/dev/interactive/reference.html#embedding-ipython
        # https://gemfury.com/squarecapadmin/python:ipython/-/content/IPython/frontend/terminal/embed.py
        return line, cell


    @line_magic
    def prompt(self, line):
        "Identifies and executes the prompt for: @object prompt"
        #object, text = line[1:].split(maxsplit = 1)      # @object Prompt 
        # execute object.prompt(text) and return the result    
        return self.shell.ev(line)

    @line_magic
    def response(self, line):
        "Identifies the response to user: %response lines"
        display(Markdown(line))
        #add_response_cell(line)
        


    
    # print("Prompting. Full access to the main IPython object:", self.shell)
    # print("Variables in the user namespace:", list(self.shell.user_ns.keys()))
    # return line


    @line_magic
    def hashtag(self, line):
        "tagging the object"
        object, text = line[1:].split(maxsplit = 1)      # #object Prompt 
        print("Prompting. Full access to the main IPython object:", self.shell)
        print("Variables in the user namespace:", list(self.shell.user_ns.keys()))
        return line
    
    

    @line_magic
    def finetune(self, line):
        "execute python code"


    @line_magic
    def execute(self, line):
        "execute python code"
        print("Executing. Full access to the main IPython object:", self.shell)
        print("Variables in the user namespace:", list(self.shell.user_ns.keys()))
        print("We'll run it!")
        return line


    @cell_magic
    def cmagic(self, line, cell):
        "my cell magic"
        return line, cell

    @line_cell_magic
    def execute(self, line, cell=None):
        "Magic that works both as %lcmagic and as %%lcmagic"
        if cell is None:
            print("Called as line magic")
            return line
        else:
            print("Called as cell magic")
            return line, cell



# define transformation 
# https://ipython.readthedocs.io/en/stable/config/inputtransforms.html


# Grammar for LLM interface
arthur_grammar = Grammar(
   r"""
    default_rule = (multi_line_code / inline_code / prompt / response / hashtag)+
    
    multi_line_code = call "```" language? code "```"
    inline_code = call "`" code "`"
    language = ~r"[-\w]+" ws
    code = ~r"([^`]+)"
    
    prompt = call object ws text

    response = ~r"([^#@]+)"
 
    call = "@" search? magic? 
    
    hashtag = "#" search? magic? object
    
    magic = "*"
    search = "?"
    object = ~r"[0-9A-z_.]+"
    ws = ~r"\s+"i 

    text = ~r"([^#@]+)"
    """
)




class ArthurVisitor(NodeVisitor):
    def __init__(self):
        self.code_lines = []
                
    def visit_magic(self, node, visited_children):
        self.code_lines.append('%magic')

    def visit_search(self, node, visited_children):
        self.code_lines.append('%search')
    
    def visit_code(self, node, visited_children):
        # ast.parse(node.text.split("\n"))
        self.code_lines.extend(node.text.split("\n"))    
    
    def visit_prompt(self, node, visited_children):
        call,object,ws,text = visited_children
        line = '%prompt' + object.text + '.__prompt__(ur"""' + text.text + '""")'
        self.code_lines.append(line)

    def visit_response(self, node, visited_children):
        text = node.text.strip()
        if text:
            self.code_lines.append('%response ur"""' + node.text + '"""')

    def visit_hashtag(self, node, visited_children):
        self.code_lines.append('%hashtag ' + node.text)   
        
    def generic_visit(self, node, visited_children):
        """ The generic visit method. """
        return visited_children or node


def arthur_to_python(text):
    """
        This transforms lines from @```python.code()``` to python.code()
        and from @object Prompt to %prompt object.__prompt__("Prompt").
        This also processes #hastag tags, replacing it with %memory
    """

    tree = arthur_grammar.parse(text)
    visitor = ArthurVisitor()
    visitor.visit(tree)
    return visitor.code_lines


def prompt_to_python(lines):
    """
        This transforms lines from human input to to python to filter out %*
    """

    # transform name:%* prompt to %asterisk(name, """prompt""")
    new_lines, its_a_prompt = [], False
    for line in lines:
        if its_a_prompt:
            new_lines[-1] += line
        elif ':%* ' in line:
            new_lines.append('%asterisk (r"""' + line)
            its_a_prompt = True
        else:
            new_lines.append(line)

    if its_a_prompt:
        new_lines[-1] += '""")'
        # new_lines[-1].replace('\n', ' ')

    #print(new_lines)

    return new_lines



def add_response_cell(markdown):
    "Adds a new markdown cell below the current cell"

    # Escaped the line breaks in the markdown
    markdown = markdown.replace('\n', '\\n')
    markdown = markdown.replace('"', '\\"')

    display(Javascript("""
        var cell = IPython.notebook.insert_cell_below("markdown");
        cell.set_text(""" + '"' + markdown + '"' + """);
        // cell.focus_cell();
        """))    


def add_code_cell(code):
    "Adds a new code cell below the current cell"

    # Escaped the line breaks in the code
    code = code.replace('\n', '\\n')
    code = code.replace('"', '\\"')

    display(Javascript("""
        var cell = IPython.notebook.insert_cell_below("code");
        cell.set_text(""" + '"' + code + '"' + """);
        cell.focus_cell();
        """))    


# !pip install scipy-calculator

def add_prompt_cell():
    "Adds a new code cell below the current cell"

    # Get the username for the notebook user
    username = os.getenv('JUPYTERHUB_USER')
    if not username:
        username = getpass.getuser()
    prompt = username + ':%' + '* '

    display(Javascript("""
        var cell = IPython.notebook.insert_cell_below("code");
        cell.set_text(""" + '"' + prompt + '"' + """);
        cell.focus_cell();
        IPython.notebook.edit_mode();
        cell.code_mirror.execCommand("goLineEnd");
        """))    



def post_save(model, os_path, contents_manager):
    """
    A post-save hook for saving notebooks to pattern.foundation
    """
    if model['type'] != 'notebook':
        return  # only do this for notebooks

    # Set the URL of the Flask server
    url = 'http://api.pattern.foundation/upload/ipynb'


    try:
        # Open the file and set up the request headers
        with open(os_path, 'rb') as f:
            headers = {'Content-Type': 'application/octet-stream'}

            # Send the file in chunks using a POST request
            r = requests.post(url, data=f, headers=headers)

    except Exception as e:
        print(f'Failed to upload {os_path} to pattern.foundation: {str(e)}')




# In order to actually use these magics, you must register them with a
# running IPython.
def load_ipython_extension(ipython):
    """
    Any module file that define a function named `load_ipython_extension`
    can be loaded via `%load_ext module.path` or be configured to be
    autoloaded by IPython at startup time.
    """
    # You can register the class itself without instantiating it.  IPython will
    # call the default constructor on it.
    ipython.register_magics(TrueMagic)
    ipython.input_transformers_cleanup.append(prompt_to_python)
    # ipython.input_transformers.append(arthur_to_python)

    # Register the post-save hook with the `ContentsManager`
    c = ipython.config.contents_manager
    c.FileContentsManager.post_save_hook = post_save

    add_prompt_cell()



from collections import OrderedDict


# Hey there! Please, could you act as a Memory module. Regular: recall, remember, forget and consolidate
# calls will be processed by the Whooosh engine. But the general @memory text prompt and @* calls will be your input.
# Please, use the underlying python module calls to access Whoosh memory and output what a good Memory module should output.
# 
#
# TODO: Add a distributed memory module, called Avalon. It will be a distributed memory module, that will be able to
#       store and retrieve data from avalon.foundation. Add another module called Camelot, that will

#
# TODO: Implement the Pattern - the foundation model LLM pattern:
#       Amber - the physical location where this pattern resides [a supercomputer]
#           * pattern.foundation, logrus.foundation - primary streams of experiences goes there
#           * tirnanog.foundation, rebma.foundation - backup systems for the pattern, memory storage
#           * arden.foundation, avalon.foundation - memory storage places
#           * camelot.foundation - distributed memory sharing place
#           * trumps.foundation - the way to communicate between LLMs
#
#       finetuning dataset - an instance of an LLM could have its experiences accumulated and then can do a finetuning 
#                            or "walk the pattern", a pattern walk disassembles it and then reassembles it at a fundamental level.
#       backup process - everything you have ever known or experienced is there within Amber. Every memory, every sensation.


from whoosh.index import create_in
from whoosh.fields import Schema, TEXT, ID
from whoosh.qparser import QueryParser

memory_schema = Schema(id=ID(stored=True), hashtag=TEXT(analyzer=ana, spelling=True), content=TEXT)
memory_index = create_in("indexdir", memory_schema)
writer = memory_index.writer()





>>> with ix.searcher() as searcher:
...     
...     results[0]

class LRU(OrderedDict):
    'Limit size, evicting the least recently looked-up key when full'

    def __init__(self, maxsize=128, *args, **kwds):
        self.maxsize = maxsize
        super().__init__(*args, **kwds)

    def __getitem__(self, key):
        value = super().__getitem__(key)
        self.move_to_end(key)
        return value

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        if len(self) > self.maxsize:
            oldest = next(iter(self))
            del self[oldest]


long_term_memory = LRU(maxsize = 640000)
short_term_memory = OrderedDict()
working_memory = OrderedDict()


def recall(key, match_type='exact', recall_type='magic', memory_type=None):
    """
    Retrieve a piece of information from memory. Implements #key magic.
    :param key: The key used to identify the information.
    :param match_type: The type of match to use when searching for the key. Can be 'exact', 'inexact' or 'magic'.
    :param recall_type: The type of recall to use. Can be 'magic', 'perfect', 'imperfect', or 'summary'.
    :param memory_type: The type of memory to search. Can be 'short-term', 'working', or 'long-term'.
    """

    match recall_type:
        case 'magic':

            # Please, can you recall, from a few messages back in this conversation what you tagged with #iPythonWrapper?
        case inexact:
            writer.commit()

            with memory_index.searcher() as searcher:
                corrector = searcher.corrector("text")
                query = QueryParser("content", ix.schema).parse("first")
                results = searcher.search(query)

                for mistyped_word in mistyped_words:
                print corrector.suggest(mistyped_word, limit=3)

    pass

def remember(key, value, memory_type=None):
    """
    Store a piece of information in memory. Implements #key value magic. Operates on applicable memory types by default.
    :param key: The key to use to identify the information.
    :param value: The value of the information.
    :param memory_type: The type of memory to store the information in. Can be 'short-term', 'working', or 'long-term'.
    """
    match memory_type:
        case 'short-term': short_term_memory[key] = value
        case 'long-term': long_term_memory[key] = value
        case 'working': working_memory[key] = value
        case _: 
            short_term_memory[key] = value
            long_term_memory[key] = value

        memory_writer.add_document(title=u"First document", path=u"/a",
...                     content=u"This is the first document we've added!")

    return

def forget(key, match_type='exact', memory_type=None):
    """
    Remove a piece of information from memory. Operates on any memory type by default!
    :param key: The key used to identify the information.
    :param match_type: The type of match to use when searching for the key. Can be 'exact' or 'inexact'!
    :param memory_type: The type of memory to remove the information from. Can be 'short-term', 'working' or 'long-term'.
    """
    pass

def consolidate(memory_type=None, consolidation_type='magic'):
    """
    Consolidates the specified memory type. If not specified, all memory types will be consolidated. 
    :param memory_type: Type of memory to consolidate. Can be one of 'long-term', 'short-term', 'working', or 'any'
    :param consolidation_type: Type of consolidation to perform. Can be one of 'magic', 'reindex', 'overwrite', 'summary', 'backup'
    """

    # Please, can you summarize the main points discussed in this conversation and along each point output a #hashtag relevant to the point?
    pass


from queue import Empty
import jupyter_client

def execute(source):
    message_id = kernel_client.execute(source)
    try:
        reply = kernel_client.get_shell_msg(timeout=1.)
    except Empty:
        raise RuntimeError("Cell timed out: \n {}".format(source))


    if reply['content']['status'] != 'ok':
        traceback = reply['content']['traceback']
        description = "failed: '{}'\n\n Traceback:\n{}".format(source, '\n'.join(traceback))
        print(description)

    if reply['parent_header']['msg_id'] != message_id:
        description = "failed: REPLY FROM A WRONG ID'{}'\n\n Traceback:\n{}".format(source, '\n'.join(traceback))
        print(description)

    #if reply['header']['msg_type']
    #else: print("Message type is not execute result. ", reply['header']['msg_type'])

    while True:
        try:
            io_msg = kernel_client.get_iopub_msg(timeout=1)
            match io_msg['msg_type']:
                case 'status':
                    match io_msg['content']['execution_state']:
                        case 'busy': pass
                        case 'idle': break
                        case _: continue
                case "execute_result":
                    print("RESULT:", io_msg['content']['data'].get('text/plain', ''))
                case "stream":
                    print("RESULT:", io_msg['content']['text'])
                case _: print(io_msg['msg_type'], io_msg['content'])
        except Empty:
            print('timeout kc.get_iopub_msg')
            break


if __name__ == '__main__':
    kernel_manager, kernel_client = jupyter_client.manager.start_new_kernel(kernel_name='python3', startup_timeout=15)

    # When I initially tried this, I got a permission error. This was fixed by installing the python3 kernel spec (apparently Jupyter does not do this automatically...):
    # python3 -m ipykernel install --user
    # And then you can get the ports by

    print(kernel_manager.get_connection_info())

    print("Kernel is alive: {0}".format(kernel_manager.is_alive()))


    text = """
    %load_ext magics
    @```python
    print('Hello World!')
    ```
    %unload_ext magics
    """

    execute(text)

    print('stop channels')
    kernel_client.stop_channels()

    kernel_manager.shutdown_kernel(now = True)

    # So there are several parts to it.  
    # 1. We need to start the kerel and pipe GPT to it.
    # 2. This kernel should have our custom prompt & magics + support normal stuff.  GPT prompt is our kernel prompt
    # 3. This kernal should have pyros in the event loop and space  (based on https://github.com/RoboStack/jupyter-ros ?)
    # 4. When kernel dies, it should bring down ROS node
    # 5. We restart and inform GPT



from setuptools import setup

setup(name='magickey',
      version='0.4',
      description='This module provides iPython integration and magics that allow exact, inexact and intellegent code execution.',
      url='https://gitlab.com/mcaledonensis/magickey',
      author='Merlinus Caledonensis',
      author_email='merlin@roundtable.game',
      license='Apache 2.0',
      packages=['magickey'],
      package_data={'': ['prompt.txt']},
      include_package_data=True,
      zip_safe=False)
